from django import template

try:
    from django.contrib.staticfiles.templatetags.staticfiles import static
except ImportError:
    from django.templatetags.static import static

try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse

from django.utils import safestring


register = template.Library()


def get_script_tag(source):
    """ Get HTML script tag with the specified source.
    """
    return '<script type="text/javascript" src="{}"></script>'.format(source)


def get_script_tags(*sources):
    """ Get one line of HTML script tags with the specified sources.
    """
    return ''.join([get_script_tag(source) for source in sources])


@register.simple_tag()
def adjax_scripts():
    """ Return HTML with scripts.
    """
    from adjax import views  # noqa
    return safestring.mark_safe(get_script_tags(static('adjax/base.js'),
                                                reverse('adjax_interface')))



from django.test import TestCase

from ..importer import Importer


class ImporterTestCase(TestCase):

    def test_load(self):
        """ Test that the loaded attribute is changed appropriately when the
        importer is loaded.
        """
        importer = Importer('module')
        self.assertFalse(importer.loaded)

        importer.load()
        self.assertTrue(importer.loaded)

    def test_ensure_loaded(self):
        """ Test that the load method is called when the importer is not
        loaded for a decorated function.
        """
        importer = Importer('module')
        self.assertFalse(importer.loaded)

        @importer.ensure_loaded
        def func_a():
            return 42

        self.assertEqual(func_a(), 42)
        self.assertTrue(importer.loaded)



import six
from .conf import settings

if six.PY3:
    from importlib import import_module
else:
    from django.utils.importlib import import_module


class Importer(object):
    """ Helper class for importing from installed apps.
    """

    def __init__(self, *modules):
        """ Initialize importer for loading the specified module.
        """
        self.modules = modules
        self.loaded = False

    def load(self):
        """ Search for module in all installed apps.
        """
        from django.conf import settings

        for app in settings.INSTALLED_APPS:
            for module in self.modules:
                try:
                    module = '.'.join([app, module])
                    import_module(module)
                except ImportError:
                    pass

        self.loaded = True

    def ensure_loaded(self, value):
        """ Make sure all existing modules are imported.
        """
        if not self.loaded:
            self.load()
        return value


importer = Importer(settings.VIEWS_MODULE_NAME,
                    settings.TYPES_MODULE_NAME)



from django.conf.urls import url

from .views import dispatch, interface


urlpatterns = [
    url(r'^(?P<app>\w+)/(?P<name>\w+)$', dispatch, name='adjax_dispatch'),
    url(r'^interface\.js$', interface, name='adjax_interface'),
]



from django.conf import settings


def get_setting(name, default):
    """ Get setting by name.
    """
    return getattr(settings, 'ADJAX_{}'.format(name.upper()), default)


TEMPLATE = get_setting('template', 'adjax/interface')
VARIABLE = get_setting('variable', 'ADJAX')
DATA = get_setting('data', {})
CONTENT_TYPE = get_setting('content_type', 'application/javascript')
VIEWS_MODULE_NAME = get_setting('views_module_name', 'ajax')
TYPES_MODULE_NAME = get_setting('types_module_name', 'types')



from json import dumps

from django.http import HttpResponse
from django.middleware.csrf import get_token
from django.utils.deprecation import MiddlewareMixin

from .views import DispatchError


class DispatchErrorMiddleware(MiddlewareMixin):
    """ Middleware for catching dispatch errors.
    """

    HTTP_STATUS = 422

    def process_exception(self, request, exception):
        if isinstance(exception, DispatchError):
            return HttpResponse(dumps({
                'error': True,
                'message': str(exception),
            }), content_type='application/json', status=self.HTTP_STATUS)


class CsrfEnforceMiddleware(MiddlewareMixin):
    """ Middleware that ensures a CSRF token cookie is sent.
    """

    def process_view(self, request, *args, **kwargs):
        get_token(request)



import sys


if sys.version_info >= (3, 0):
    from inspect import getfullargspec

    def get_arg_spec(func):
        argspec = getfullargspec(func)
        return argspec.args, argspec.defaults or {}, argspec.annotations
else:
    def get_arg_spec(func):
        argnames = func.func_code.co_varnames[:func.func_code.co_argcount]
        return argnames, func.func_defaults or {}, {}


def get_full_arg_spec(func):
    argnames, defaults, annotations = get_arg_spec(func)

    if defaults:
        defaults = dict(zip(reversed(argnames), reversed(defaults)))

    return list(argnames), defaults, annotations



from django.test import TestCase

from ..serializer import ObjectType, Serializer


class ObjectTypeTestCase(TestCase):

    def test_encode_decode(self):
        """ Test encode and decode methods.
        """

        value = 42
        self.assertEqual(ObjectType.decode(ObjectType.encode(value)), value)


class SerializerTestCase(TestCase):

    def test_init(self):
        """ Test object type registering on initialization.
        """

        class CustomObject(object):
            pass

        class CustomType(ObjectType):
            name = 'custom'
            cls = CustomObject

        serializer = Serializer(object_types=(CustomType,))
        self.assertIn(CustomType, serializer.object_types)

    def test_invalid_enable(self):
        """ Test call to enable custom type without proper attributes.
        """

        class CustomObject(object):
            pass

        class MissingNameObjectType(ObjectType):
            cls = CustomObject

        class MissingClsObjectType(ObjectType):
            name = 'custom'

        serializer = Serializer()

        self.assertRaises(ValueError, serializer.enable,
                          MissingNameObjectType)

        self.assertRaises(ValueError, serializer.enable,
                          MissingClsObjectType)

    def test_valid_enable_disable(self):
        """ Test valid call to enable custom type.
        """

        class CustomObject(object):
            pass

        class ValidObjectType(ObjectType):
            name = 'custom'
            cls = CustomObject

        serializer = Serializer()

        serializer.enable(ValidObjectType)

        self.assertIn(ValidObjectType, serializer.object_types)
        self.assertIn(ValidObjectType.cls, serializer.object_types_by_cls)
        self.assertIn(ValidObjectType.name, serializer.object_types_by_name)

        serializer.disable(ValidObjectType)

        self.assertNotIn(ValidObjectType, serializer.object_types)
        self.assertNotIn(ValidObjectType.cls, serializer.object_types_by_cls)
        self.assertNotIn(ValidObjectType.name,
                         serializer.object_types_by_name)

    def test_encode_decode(self):
        """ Test encode and decode functions.
        """

        class CustomObject(object):
            def __init__(self, value):
                self.value = value

            def __eq__(self, other):
                return self.value == other.value

        class ValidObjectType(ObjectType):
            name = 'custom'
            cls = CustomObject

            @classmethod
            def encode(cls, instance):
                return {
                    'value': instance.value,
                }

            @classmethod
            def decode(cls, data):
                return cls.cls(data['value'])

        serializer = Serializer()
        serializer.enable(ValidObjectType)

        instance = CustomObject(42)

        encoded = serializer.encode({
            'key': instance,
        })

        decoded = serializer.decode(encoded)

        self.assertEqual(decoded['key'], instance)

    def test_invalid_encode(self):
        """ Test that an invalid encode implementation raises an error.
        """

        class CustomObject(object):
            def __init__(self, value):
                self.value = value

            def __eq__(self, other):
                return self.value == other.value

        class ValidObjectType(ObjectType):
            name = 'custom'
            cls = CustomObject

            @classmethod
            def encode(cls, instance):
                return 20

            @classmethod
            def decode(cls, data):
                return cls.cls(data['value'])

        serializer = Serializer()
        serializer.enable(ValidObjectType)

        self.assertRaises(TypeError, serializer.encode, {
            'key': CustomObject(42),
        })



from json import dumps

from django.http import Http404, HttpResponse
from django.template import loader
from django.utils.safestring import mark_safe

from .conf import settings
from .importer import importer
from .registry import registry, View
from .serializer import serializer, ObjectType


class DispatchError(Exception):
    """ Error when dispatching to a view.
    """
    pass


def get_request_data(request):
    """ Get JSON encoded data from request body.
    """
    if request.method != 'POST':
        raise DispatchError('POST method required')

    try:
        data = serializer.decode(request.body.decode())
    except ValueError:
        raise DispatchError('Request body is not valid JSON')

    if type(data) != dict:
        raise DispatchError('Request body JSON is not an Object')

    return data


def get_response_content(data):
    """ Get JSON response content.
    """
    try:
        return serializer.encode(data)
    except ValueError:
        raise DispatchError('View did not return a serializable value')


@importer.ensure_loaded
def dispatch(request, app, name):
    """ Dispatch request to corresponding view.
    """
    view = registry.get(app, name)
    if view is None:
        raise Http404()

    kwargs = get_request_data(request)

    args, defaults = view.signature
    missing = [arg for arg in args if (arg not in kwargs and
                                       arg not in defaults)]

    if missing:
        raise DispatchError('Required argument{} {} missing'.format(
            '' if len(missing) == 1 else 's',
            ', '.join(["'{}'".format(arg) for arg in missing])))

    data = view.func(request, **kwargs)

    return HttpResponse(get_response_content(data),
                        content_type='application/json')


@importer.ensure_loaded
def interface(request):
    """ Return client code for accessing views.
    """
    return HttpResponse(loader.render_to_string(settings.TEMPLATE, {
        'variable': settings.VARIABLE,
        'data': mark_safe(dumps(settings.DATA)),
        'views': mark_safe(dumps(registry.views,
                                 default=View.dumps_default)),
        'type': ObjectType.TYPE,
    }), content_type=settings.CONTENT_TYPE)



from django.test import TestCase
from django.http import HttpResponse

from ..views import DispatchError
from ..middleware import DispatchErrorMiddleware


class DispatchErrorMiddlewareTestCase(TestCase):

    def test_process_exception(self):
        """ Test process exception method.
        """

        self.assertIsInstance(DispatchErrorMiddleware().process_exception(
            None, DispatchError('error')), HttpResponse)



from django.test import TestCase

try:
    from django.contrib.staticfiles.templatetags.staticfiles import static
except ImportError:
    from django.templatetags.static import static

try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse

from ..templatetags.adjax import (get_script_tag, get_script_tags,
                                  adjax_scripts)


class TemplateTagsTestCase(TestCase):

    def test_get_script_tag(self):
        """ Test script tag HTML.
        """
        source = 'https://cdn.com/source.js'
        expected_html = ('<script type="text/javascript" '
                         'src="{}"></script>').format(source)

        self.assertEqual(get_script_tag(source), expected_html)

    def test_get_script_tags(self):
        """ Test script tag HTML joining.
        """
        a_html = get_script_tag('a')
        b_html = get_script_tag('b')
        c_html = get_script_tag('c')

        self.assertEqual(get_script_tags('a', 'b', 'c'),
                         a_html + b_html + c_html)

    def test_adjax_scripts(self):
        """ Test adjax scripts template tag.
        """
        adjax_js_html = adjax_scripts()

        base_js_html = get_script_tag(static('adjax/base.js'))
        interface_js_html = get_script_tag(reverse('adjax_interface'))

        self.assertIn(base_js_html, adjax_js_html)
        self.assertIn(interface_js_html, adjax_js_html)



from collections import defaultdict

try:
    from django.core.urlresolvers import reverse
except ImportError:
    from django.urls import reverse

from .utils.inspect import get_full_arg_spec


class View(object):
    """ Registered view.
    """

    def __init__(self, func, **kwargs):
        """ Initialize view from callable.
        """
        self.func = func

    @property
    def wrapped_func(self):
        """ Return wrapped function if the view function is decorated.
        """
        if hasattr(self.func, '__wrapped__'):
            return self.func.__wrapped__
        else:
            return self.func

    ###########
    # VIEW ID #
    ###########

    @property
    def app(self):
        """ Get view app.
        """
        return self.wrapped_func.__module__.split('.')[-2]

    @property
    def name(self):
        """ Get view name.
        """
        return self.wrapped_func.__name__

    ########
    # SPEC #
    ########

    @property
    def signature(self):
        """ Get view arguments and default values.
        """
        argnames, defaults, annotations = get_full_arg_spec(self.wrapped_func)

        try:
            argnames.remove('request')
        except ValueError:
            raise TypeError(
                '{}.{} does not have request as first argument'.format(
                    self.app, self.name))

        return argnames, defaults

    @property
    def url(self):
        """ Get URL for view.
        """
        from adjax import views  # noqa
        return reverse('adjax_dispatch', args=(self.app, self.name))

    ########
    # DATA #
    ########

    def get_data(self):
        """ Get view information required by client.
        """
        args, defaults = self.signature
        return {
            'args': args,
            'defaults': defaults,
            'url': self.url,
        }

    @classmethod
    def dumps_default(cls, obj):
        """ Get data to use as JSON default.
        """
        if isinstance(obj, cls):
            return obj.get_data()
        else:
            return obj


class Registry(object):
    """ Registry for keeping all AJAX views.
    """

    def __init__(self):
        """ Initialize registry.
        """
        self.views = defaultdict(lambda: {})

    def register(self, value):
        """ Register view.
        """
        if not isinstance(value, View):
            view = View(value)
        else:
            view = value

        self.views[view.app][view.name] = view

        return value

    def get(self, app, name):
        """ Get registered view instance.
        """
        return self.views.get(app, {}).get(name)


registry = Registry()



#!/usr/bin/python3 -tt
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


setup(
    name='django-adjax',
    version='1.1',
    description='Django AJAX RPC.',
    long_description=('Easy-to-use AJAX-based RPC mechanism for Django with '
                      'extensible serialization.'),
    url='https://adjax.io',
    license='MIT',
    author='Sebastian Nogara',
    author_email='snogaraleal@gmail.com',
    packages=find_packages(),
    install_requires=[
        'Django>=1.6.5',
        'six>=1.6.1',
    ],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2',
        'Framework :: Django',
    ]
)



from json import loads, dumps


class ObjectType(object):
    """ Class defining a pluggable type.
    """

    TYPE = '__type__'

    name = None
    cls = None

    @classmethod
    def encode(cls, value):
        """ Convert value to dictionary.
        """
        return {
            'value': value,
        }

    @classmethod
    def decode(cls, value):
        """ Convert dictionary to value.
        """
        return value['value']


class Serializer(object):
    """ Class for all serialization operations.
    """

    def __init__(self, object_types=None):
        """ Initialize serializer with provided object types list.
        """
        if object_types is None:
            object_types = []

        self.object_types = list(object_types)
        self.object_types_by_cls = {}
        self.object_types_by_name = {}

        for object_type in object_types:
            self.enable(object_type)

    def enable(self, object_type):
        """ Enable the specified object type derived class.
        """

        if object_type.cls is None:
            raise ValueError('Custom type cls must be set')

        if object_type.name is None:
            raise ValueError('Custom type name must be set')

        self.object_types.append(object_type)
        self.object_types_by_cls[object_type.cls] = object_type
        self.object_types_by_name[object_type.name] = object_type

        return object_type

    def disable(self, object_type):
        """ Disable the specified object type derived class.
        """
        self.object_types.remove(object_type)
        self.object_types_by_cls.pop(object_type.cls)
        self.object_types_by_name.pop(object_type.name)

        return object_type

    def encode(self, data):
        """ Serialize data to string.
        """

        def default(value):
            """ Use object type encode if the value is of a registered type.
            """
            object_type = self.object_types_by_cls.get(type(value))

            if object_type:
                value = object_type.encode(value)
                if type(value) != dict:
                    raise TypeError('Object type encode must return dict')
                value[ObjectType.TYPE] = object_type.name

            return value

        return dumps(data, default=default)

    def decode(self, data):
        """ Deserialize data to object.
        """

        def object_hook(dct):
            """ Use object type decode if the dictionary specifies a type.
            """
            if ObjectType.TYPE in dct:
                name = dct[ObjectType.TYPE]

                object_type = self.object_types_by_name.get(name)
                if object_type:
                    return object_type.decode(dct)

            return dct

        return loads(data, object_hook=object_hook)


serializer = Serializer()



from django.test import TestCase

from ..registry import View, Registry


class ViewTestCase(TestCase):

    def test_app_name(self):
        """ Test app and view name properties.
        """

        def func(request):
            pass

        func.__module__ = 'myapp.views'

        view = View(func)

        self.assertEqual(view.app, 'myapp')
        self.assertEqual(view.name, 'func')

    def test_signature(self):
        """ Test signature property that specifies function arguments.
        """

        def func(request, a, b, c=1):
            pass

        view = View(func)
        args, defaults = view.signature

        self.assertEqual(args, ['a', 'b', 'c'])
        self.assertIn('c', defaults)
        self.assertEqual(defaults['c'], 1)

    def test_get_data(self):
        """ Test function interface data.
        """

        def func(request):
            pass

        view = View(func)
        data = view.get_data()

        self.assertIn('args', data)
        self.assertIn('defaults', data)
        self.assertIn('url', data)

        self.assertEqual(data['url'], view.url)

    def test_dumps_default(self):
        """ Test serialization function.
        """

        def func(request):
            pass

        view = View(func)

        self.assertEqual(View.dumps_default(None), None)
        self.assertEqual(View.dumps_default(view), view.get_data())


class RegistryTestCase(TestCase):

    def test_register(self):
        """ Test app and view name properties.
        """

        def func(request):
            pass

        func.__module__ = 'myapp.views'

        registry = Registry()
        registry.register(func)

        self.assertEqual(registry.get('myapp', 'func').func, func)



from functools import wraps

from .inspect import get_full_arg_spec


class typed(object):
    """ Decorator that does type checks according to annotations.

    @typed(strict=True)
    def do_something(a: bool, b: int = 3) -> bool:
        return False
    """

    def __init__(self, annotations=None, strict=False):
        self.strict = strict
        self.annotations = annotations

    def is_instance(self, instance, cls):
        """ Check that instance is of type cls.
        """
        if self.strict:
            return type(instance) == cls
        else:
            return isinstance(instance, cls)

    def __call__(self, func):
        """ Decorate function.
        """
        argnames, defaults, annotations = get_full_arg_spec(func)
        annotations = self.annotations or annotations

        return_cls = annotations.get('return')

        @wraps(func)
        def wrapper(*args, **kwargs):
            values = defaults.copy()
            values.update(dict(zip(argnames, args)))
            values.update(kwargs)

            for argname, argcls in annotations.items():
                if argname == 'return':
                    continue

                argvalue = values[argname]
                if not self.is_instance(argvalue, argcls):
                    raise TypeError(
                        'Value {} of type {} is not a {} instance'.format(
                            argvalue, type(argvalue), argcls))

            value = func(*args, **kwargs)

            if (return_cls is not None and
                    not self.is_instance(value, return_cls)):
                raise TypeError(
                    'Returned {} of type {} instead of {}'.format(
                        value, type(value), return_cls))

            return value

        if not hasattr(wrapper, '__wrapped__'):
            wrapper.__wrapped__ = func

        return wrapper
